//-*-C++-*- Header file for the Catalog Class
// Copyright (C) 1997 by John P. Weiss
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the Artistic License, included as the file
// "LICENSE" in the source code archive.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//
// You should have received a copy of the file "LICENSE", containing
// the License John Weiss originally placed this program under.
//
// This file is unreadable, because the comment-docifier is
// brainedead.  Try a "make srcdocs" to generate the
// HTML-documentation.
//
#ifndef _Catalog_h_
#define _Catalog_h_
#include <stdio.h>


// The default number of frames per second on a CD.
#define FRAME_PER_SEC 75
#define MAXCHARLEN 256


/**
  This is a class to contain a CD or playlist's information for a
  catalog database.  It is only a container; it does nothing towards
  database maintainence or modification.

  {\bf NOTE 1:} All class methods are inlined and defined after the
  class declaration.  Please be sure you compile accordingly!

  {\bf NOTE 2:} Methods that access protected arrays {\em do not}
  check those indices to see if they are in range!  That's your job as
  the programmer.

  The magic number is calculated using Ti Kan's magic number
  algorithm, as presented in the source code for "wm2xmcd", part of
  his "xmcd" CD player program.  See a source code distribution for
  "xmcd" for more info; direct all queries about the magic number to
  ti@amb.org.

  {\bf
  \begin{verbatim}
  
  ----------------------------------------------------------
  ----------------------------------------------------------
  \end{verbatim}
  }

  @memo Information for a CD or list of audio files.
  @see AudDB, AudioDev
  */
class Catalog
{
   public:
	 /**
	   List of indices used to create a playlist.
	   The numbers in {\bf playlist} are used by \Ref{plTime}() and
	   \Ref{plName}() to index the list of track times and names.  If
	   {\tt playlist[i]>=\Ref{len}()} anywhere in the list, you'll get a
	   segmentation fault.

	   {\bf Do not} deallocate this array; you'll cause
	   \Ref{~Catalog}() to crash.
	   @memo
	   */
	 int *playlist;
	 /**
	   Length of \Ref{playlist}.
	   @memo
	   */
	 int pLen;

	 /**
	   Constructor for the {\bf Catalog} class.
	   
	   The constructor initializes several internal arrays and
	   strings.  These arrays must have some non-zero length,
	   specified by {\em len0}.  The default is set to 2, to prevent
	   the destructor from segfaulting.

	   The other two parameters are not quite as critical.  {\em
	   units} specifies the unit of time you're using in this {\bf
	   Catalog}.  It defaults to units of frames [common on CD's]; if
	   you wanted to make a {\bf Catalog} of sound files, you'd set
	   {\em units} to the number of bytes per second in each file.

	   The initial names of each item are the string {\em tag} plus
	   the item number.
	   @memo
	   */
	 Catalog(unsigned int len0=2,
		    unsigned int units=FRAME_PER_SEC,
		    char *tag="Track");

	 /**
	   Copy Constructor for the {\bf Catalog} class.
	   @memo
	   */
	 Catalog(Catalog& that);

	 /**
	   Destructor for the {\bf Catalog} class.
	   @memo
	   */
	 ~Catalog(void);

	 /**
	   @return The number of items in this {\bf Catalog}.
	   @memo
	   */
	 int len(void);

	 /**
	   @return The title of this {\bf Catalog}.
	   @memo
	   */
	 const char *title(void);

	 /**
	   @return The artist/author of whatever this {\bf Catalog}
	   describes [CD, .midi file, etc.].
	   @memo
	   */
	 const char *artist(void);

	 /**
	   @return The label of the {\em idx}-th item.
	   @memo
	   */
	 const char *name(unsigned int idx);

	 /**
	   @return The time value of the {\em idx}-th item.
	   @see setTime
	   @memo
	   */
	 int time(unsigned int idx);

	 /**
	   Equivalent to \Ref{name}(\Ref{playlist}[{\em idx}]).
	   @memo
	   */
	 char *plName(unsigned int idx);

	 /**
	   Equivalent to \Ref{time}(\Ref{playlist}[{\em idx}]).
	   @memo
	   */
	 int plTime(unsigned int idx);

	 /**
	   Resets the \Ref{playlist} array to its initial values.  Useful
	   if you've made lots of changes and want \Ref{plName}() and
	   \Ref{plTime}() to start accessing items in numerical order
	   again...
	   @memo
	   */
	 void resetPlaylist(void);

	 /**
	   @return The unique magic number for this {\bf Catalog}, or
	   zero if \Ref{setTime}() or \Ref{genMagic}() have never been
	   called.
	   @memo
	   */
	 unsigned long magic(void);

	 /**
	   Sets the time value of the {\em idx}-th item to {\em val}.

	   Each item has associated with it a unique time-index, which
	   typically specifies its position on the CD.  If you are
	   cataloging a list of soundfiles, you should set {\em val} to
	   the \Ref{time}() of the previous item plus the length of that
	   item.  

	   {\em val} should always be in the units specifed when you
	   called the constructor [{\it i.e.} frames per second, bytes
	   per second, etc...].

	   This method automatically calls \Ref{genMagic} to update the
	   {\bf Catalog} magic-number.
	   @memo
	   */
	 void setTime(unsigned int idx, int val);

	 /**
	   Sets the label of the {\em idx}-th item to a copy of the string
	   specified in {\em ostr}.
	   @memo
	   */
	 void setName(unsigned int idx, const char *ostr);

	 /**
	   Sets the {\bf Catalog} title to a copy of the string specified
	   in {\em ostr}.  
	   @memo
	   */
	 void setTitle(const char *ostr);

	 /**
	   Sets the {\bf Catalog} artist to a copy of the string specified
	   in {\em ostr}.  
	   @memo
	   */
	 void setArtist(const char *ostr);

	 /**
	   Generates the magic number for this {\bf Catalog} using the
	   times given in \Ref{cTimes}.

	   Based on the magic number generator by Ti Kan [ti@amb.org];
	   please direct all questions about it to him.
	   @memo
	   */
	 void genMagic(void);

	 /**
	   Overloaded assignment operator.  Copies the arrays instead of
	   copying their pointers.
	   @memo
	   */
	 Catalog& operator=(Catalog& that);

	 /**
	   Is {\em true} if both {bf Catalog}s have the same magic number.
	   @memo
	   */
	 bool operator==(Catalog that);

	 /**
	   Is {\em true} if both {bf Catalog}s don't have the same magic
	   number.
	   @memo
	   */
	 bool operator!=(Catalog that);

	 /**
	   An exceedingly arcane way to set the magic number by hand if
	   you don't have the list of time-indices.  Calling
	   \Ref{genMagic} will undo any effects of this operator.

	   @see setTime
	   @memo
	   */
	 void operator<<=(int newmagick);

   protected:
	 //
	 int cLen;
	 int uPerSec;
	 unsigned long cMagic;
	 char **cNames;
	 int *cTimes;
	 char *cTitle;
	 char *cArtist;
	 bool nomagic;
	 //
	 int magickChecksum(int i);
	 void copystring_(char *dstr, const char *cstr, int len);
	 void copyFrom_(Catalog& that);
};



//|||||||||||||||||||||
//
// End Class Defn.
// Begin Method Defns.
//
//|||||||||||||||||||||


// Generates the checksum used by \Ref{genMagic}.
// Based on the magic number generator by Ti Kan [ti@amb.org]; please
// direct all questions about it to him.
inline
int Catalog::magickChecksum( int i )
{
   char b[12];
   char *p;
   int out=0;
   sprintf(b,"%lu",(unsigned long)i);
   for(p=b; *p != '\0'; p++)
	 out += (*p - '0');
   return out;
}


// Generates the magic number for this {\bf Catalog} using the times
// given in \Ref{cTimes}.
// Based on the magic number generator by Ti Kan [ti@amb.org]; please
// direct all questions about it to him.
inline
void Catalog::genMagic( void )
{
   int i,tt=0,cs=0;

   for(i=0; i<cLen; i++)
	 {
	 cs += magickChecksum(cTimes[i]/uPerSec);
	 tt += cTimes[i+1]/uPerSec - cTimes[i]/uPerSec;
	 }
   cMagic = ((cs % 0xff) << 24 | tt << 8 | cLen);
   nomagic = false;
}


//-------------------------------------------------------------------------
// Remaining code is once again mine. -- jpw
//-------------------------------------------------------------------------


// Internal string copying function.  I was having problems with
// strncpy, so I decided to write something I'd have comtrol over.
inline
void Catalog::copystring_(char *dstr, const char *cstr, int len)
{
   int j;

   j=0;
   while( (j<len) && (cstr[j] != '\0') )
	 {
	 dstr[j] = cstr[j];
	 j++;
	 }
   if(j<len)
	 dstr[j] = '\0';
   else
	 dstr[len-1] = '\0';
}


// Used by copy constructor and assigment operators, so it needs to be
// declared first.
//
inline
void Catalog::copyFrom_( Catalog& that )
{
   int i,oldlen;
   
   oldlen = cLen;
   cLen = that.cLen;
   pLen = that.pLen;
   uPerSec = that.uPerSec;
   // Make sure it has a magic number before we do the assign.
   if(that.nomagic)
	 that.genMagic();
   cMagic = that.cMagic;
   copystring_(cTitle,that.cTitle,MAXCHARLEN);
   copystring_(cArtist,that.cArtist,MAXCHARLEN);

   delete [] cTimes;
   for(i=0; i<oldlen; i++)
	 delete [] cNames[i];
   delete [] cNames;
   delete [] playlist;
   cNames = new char *[cLen];
   cTimes = new int[cLen+1];
   playlist = new int[cLen];

   for(i=0; i<cLen; i++)
	 {
	 cTimes[i] = that.cTimes[i];
	 cNames[i] = new char [MAXCHARLEN];
	 copystring_(cNames[i],that.cNames[i],MAXCHARLEN);
	 }
   cTimes[cLen] = that.cTimes[cLen];

   for(i=0; i<pLen; i++)
	 {
	 playlist[i] = that.playlist[i];
	 }
}


// Constructors:
//
inline
Catalog::Catalog( unsigned int len0, unsigned int units, char *tag )
{
   int i;

   if(len0 <= 0)
	 len0 = 2;

   cLen = pLen = len0;
   uPerSec = units;
   cMagic = 0L;
   nomagic = true;
   cTitle = new char[MAXCHARLEN];
   cArtist = new char[MAXCHARLEN];
   cNames = new char *[cLen];
   cTimes = new int[cLen+1];
   playlist = new int[cLen];
   for(i=0; i<cLen; i++)
	 {
	 cNames[i] = new char[MAXCHARLEN];
	 cTimes[i] = 0;
	 playlist[i] = i;
	 sprintf(cNames[i],"%s %d",tag,i+1);
	 }
   cArtist[0] = 0;
   cTitle[0] = 0;
}


// Copy constructor
//
inline
Catalog::Catalog( Catalog& that )
{
   copyFrom_(that);
}


// Destructor
//
inline
Catalog::~Catalog( void )
{
   int i;

   delete [] playlist;
   delete [] cTimes;
   delete [] cTitle;
   delete [] cArtist;
   for(i=0; i<cLen; i++)
	 delete [] cNames[i];
   delete [] cNames;   
}


inline
int Catalog::len( void )
{
   return cLen;
}


inline
int Catalog::time(unsigned int idx)
{
   return cTimes[idx];
}


inline
const char *Catalog::title(void)
{
   return cTitle;
}


inline
const char *Catalog::artist(void)
{
   return cArtist;
}


inline
const char *Catalog::name(unsigned int idx)
{
   return cNames[idx];
}


inline
int Catalog::plTime(unsigned int idx)
{
   return cTimes[playlist[idx]];
}


inline
char *Catalog::plName(unsigned int idx)
{
   return cNames[playlist[idx]];
}


inline
unsigned long Catalog::magic( void )
{
   return cMagic;
}


inline
void Catalog::resetPlaylist(void)
{
   int i;
   if( (pLen > 0) && (playlist != NULL) )
	 delete [] playlist;
   pLen = cLen;
   playlist = new int[cLen];
   for(i=0; i<cLen; i++)
	 {
	 playlist[i] = i;
	 }
}


inline
void Catalog::setTime(unsigned int idx, int val)
{
   cTimes[idx] = val;
   genMagic();  //Always recompute the magic number after a change.
}


inline
void Catalog::setName(unsigned int idx, const char *ostr)
{
   copystring_(cNames[idx],ostr,MAXCHARLEN);
}


inline
void Catalog::setTitle(const char *ostr)
{
   copystring_(cTitle,ostr,MAXCHARLEN);
}


inline
void Catalog::setArtist(const char *ostr)
{
   copystring_(cArtist,ostr,MAXCHARLEN);
}


inline
Catalog& Catalog::operator=( Catalog& that )
{
   copyFrom_(that);
   return (*this);
}


inline
bool Catalog::operator==( Catalog that )
{
   // Make sure we have magic numbers before we compare....
   if(that.nomagic)
	 that.genMagic();
   if(nomagic)
	 genMagic();

   return (cMagic == that.cMagic);
}


inline
bool Catalog::operator!=( Catalog that )
{
   // Make sure we have magic numbers before we compare....
   if(that.nomagic)
	 that.genMagic();
   if(nomagic)
	 genMagic();

   return (cMagic != that.cMagic);
}


// This is an arcane way of specifying the magic number externally.
inline
void Catalog::operator<<=( int newmagick )
{
   cMagic = newmagick;
   nomagic = false;
}




#endif // _Catalog_h_
//||||||||||||||||||||||||||||||||
//
// End
// 
